#include "Demo.h"
#include "game_over.h"

/** VARIABLES */

// Paddle move and game_over
static uint16_t old_cmd = MOV_POSITION_10 + 1;
static uint8_t life_cnt = DEF_NUM_LIVES;

// Score variables
static uint16_t overall_score;
static uint8_t bricks_destroyed = DIF_LEVEL_1;
static uint8_t score_multiplier = DIF_MULTIPLIER_1; 

/** FUNCTIONS */

/* PRIVATE_FUNCTIONS */

// RGB Matrix
static void drv_writePixel( uint16_t x, uint16_t y, uint16_t color )
{
    matrixrgb_writePixel( x, y, color );
    /* Space for potential delays */
}

static void drv_paintObject( demo_object_t object, uint8_t width, uint8_t height )
{
    uint16_t y;
    uint16_t x;

    for( y = object.y; y < object.y + height; y++ )
        for( x = object.x; x < object.x + width; x++ )
            drv_writePixel( x, y, object.color );
}

// Score related
static void drv_updateScore( )
{   
    bricks_destroyed++;

    if( bricks_destroyed >= DIF_LEVEL_3 )
        if( bricks_destroyed >= DIF_LEVEL_4 )
            if( bricks_destroyed >= DIF_LEVEL_5 )
            {
                switch ( bricks_destroyed )
                {
                    case DIF_BRONZE:
                        score_multiplier = DIF_MULTIPLIER_BRONZE;
                        break;
                    case DIF_SILVER:
                        score_multiplier = DIF_MULTIPLIER_SILVER;
                        break;
                    case DIF_GOLD:
                        score_multiplier = DIF_MULTIPLIER_GOLD;
                        break;    
                    default:
                        score_multiplier = DIF_MULTIPLIER_5;
                }
            }
            else
                score_multiplier = DIF_MULTIPLIER_4;
        else
            score_multiplier = DIF_MULTIPLIER_3;
    else
        score_multiplier = DIF_MULTIPLIER_2;
        
    overall_score += SCORE_BRICK_VAL * score_multiplier;
}

// Ball
static uint8_t drv_determineBrickCollision( demo_ball_t ball, demo_bounds_t bounds )
{
    if( ball.x == bounds.x_min )
        if( ball.y == bounds.y_min )
            return CASE_UP_LEFT;
        else
            if( ball.y == bounds.y_max )
                return CASE_DOWN_LEFT;
            else
                return CASE_LEFT;
    else if( ball.x == bounds.x_max )
        if( ball.y == bounds.y_min )
            return CASE_UP_RIGHT;
        else
            if( ball.y == bounds.y_max )
                return CASE_DOWN_RIGHT;
            else
                return CASE_RIGHT;
    else
        if( ball.y == bounds.y_max )
            return CASE_DOWN;
        else
            return CASE_UP; 
}

static uint8_t drv_redirectBall( uint8_t direction )
{
    switch( direction )
    {
        /* Vertical bounce directions + diagonal edge bounce directions */

        // Vertical cases
        case DIR_NORTH:
            return RANDOM( DIR_SOUTHEAST, DIR_SOUTHWEST );
        case DIR_SOUTH:
            return ( RANDOM( DIR_NORTH, DIR_EAST ) + 7 ) % 8;
        // Edge cases
        case DIR_CEILING_1:
            return DIR_SOUTHEAST;
        case DIR_CEILING_2:
            return DIR_SOUTHWEST;
        case DIR_RIGHT_EDGE_1:
            return DIR_NORTHWEST;
        case DIR_RIGHT_EDGE_2:
            return DIR_SOUTHWEST;
        case DIR_PADDLE_1:
            return DIR_NORTHEAST;
        case DIR_PADDLE_2:
            return DIR_NORTHWEST;
        case DIR_LEFT_EDGE_1:
            return DIR_SOUTHEAST;
        case DIR_LEFT_EDGE_2:
            return DIR_NORTHEAST;

        /* Diagonal brick bounce directions */

        // Side cases
        case CASE_UP:
            if( direction == DIR_SOUTHEAST )
                return DIR_NORTHEAST;
            else 
                return DIR_NORTHWEST;
        case CASE_RIGHT:
            if( direction == DIR_NORTHWEST )
                return DIR_NORTHEAST;
            else 
                return DIR_SOUTHEAST;
        case CASE_DOWN:
            if( direction == DIR_NORTHEAST )
                return DIR_SOUTHEAST;
            else
                return DIR_SOUTHWEST;
        case CASE_LEFT:
            if( direction == DIR_SOUTHEAST )
                return DIR_SOUTHWEST;
            else
                return DIR_NORTHWEST;
        // Angle cases
        case CASE_UP_RIGHT:
            return DIR_NORTHEAST; 
        case CASE_DOWN_RIGHT:
            return DIR_SOUTHEAST;
        case CASE_DOWN_LEFT:
            return DIR_SOUTHWEST;
        case CASE_UP_LEFT:
            return DIR_NORTHWEST;
    }
}

static void drv_repositionBall( demo_ball_t *ball )
{
    switch( ball->direction )
    {
        case 0:
            ball->y -= 1;
            break;
        case 1:
            ball->x += 1;
            ball->y -= 1;
            break;
        case 2:
            ball->x += 1;
            break;
        case 3:
            ball->x += 1;
            ball->y += 1;
            break;
        case 4:
            ball->y += 1;
            break;
        case 5:
            ball->x -= 1;
            ball->y += 1;
            break;
        case 6:
            ball->x -= 1;
            break;
        case 7:
            ball->x -= 1;
            ball->y -= 1;
    }
}

// Collision check
static uint8_t drv_edgeCheck( demo_ball_t *ball )
{
    uint16_t topLeftLimit = PANEL_SIZE_MIN - 1;
    uint16_t botRightLimit = PANEL_SIZE_MAX + 1;

    if( ball->x == topLeftLimit )
    {
        if( ball->direction == DIR_SOUTHWEST )
            ball->direction = drv_redirectBall( DIR_LEFT_EDGE_1 );
        else if( ball->direction == DIR_NORTHWEST )
            ball->direction = drv_redirectBall( DIR_LEFT_EDGE_2 );
        else
            ball->direction = drv_redirectBall( ball->direction );

        return STATUS_HIT_LEFT_EDGE;
    }
    if( ball->y == topLeftLimit )
    {
        if( ball->direction == DIR_NORTHEAST )
            ball->direction = drv_redirectBall( DIR_CEILING_1 );
        else if( ball->direction == DIR_NORTHWEST )
            ball->direction = drv_redirectBall( DIR_CEILING_2 );
        else
            ball->direction = drv_redirectBall( ball->direction );

        return STATUS_HIT_CEILING;
    }
    if( ball->x == botRightLimit )
    {
        if( ball->direction == DIR_NORTHEAST )
            ball->direction = drv_redirectBall( DIR_RIGHT_EDGE_1 );
        else if( ball->direction == DIR_SOUTHEAST )
            ball->direction = drv_redirectBall( DIR_RIGHT_EDGE_2 );
        else
            ball->direction = drv_redirectBall( ball->direction );

        return STATUS_HIT_RIGHT_EDGE;
    }
    if( ball->y == botRightLimit )
    {
         return STATUS_HIT_FLOOR;
    }

    return STATUS_NO_CONTACT;
}

static uint8_t drv_paddleCheck( demo_ball_t *ball, demo_object_t *paddle )
{
    if( paddle->y <= ball->y )
        if( paddle->y + PAD_DEF_HEIGHT - 1 >= ball->y )
            if( paddle->x <= ball->x )
                if( paddle->x + PAD_DEF_WIDTH - 1 >= ball->x )
                {
                    if( ball->direction == DIR_SOUTHEAST )
                        ball->direction = drv_redirectBall( DIR_PADDLE_1 );
                    else if( ball->direction == DIR_SOUTHWEST )
                        ball->direction = drv_redirectBall( DIR_PADDLE_2 );
                    else
                        ball->direction = drv_redirectBall( ball->direction );

                    return STATUS_HIT_PADDLE;
                }

    return STATUS_NO_CONTACT;
}

static uint8_t drv_brickCheck( demo_ball_t *ball, demo_object_t *wall )
{
    demo_bounds_t bounds;

    uint8_t max = WALL_SIZE;
    uint8_t min = 0;
    uint8_t index = max / 2;
    uint8_t old_index = 0;

    while( min <= index && index <= max && index != old_index )
    {
        if( wall[ index ].y <= ball->y )
            if( ball->y <= wall[ index ].y + BRICK_DEF_HEIGHT - 1 )
                if( wall[ index ].x <= ball->x )
                    if( ball->x <= wall[ index ].x + BRICK_DEF_WIDTH - 1 )
                        if( wall[ index ].color == COLOR_BLACK )
                            return STATUS_NO_CONTACT;
                        else
                        {
                            bounds.x_min = wall[ index ].x;
                            bounds.y_min = wall[ index ].y;
                            bounds.x_max = wall[ index ].x + BRICK_DEF_WIDTH - 1;
                            bounds.y_max = wall[ index ].y + BRICK_DEF_HEIGHT - 1;

                            ball->direction = drv_redirectBall( drv_determineBrickCollision(
                                                              ( *ball ), bounds ) );
                            wall[ index ].color = COLOR_BLACK;
                            drv_paintObject( wall[ index ], BRICK_DEF_WIDTH, BRICK_DEF_HEIGHT );

                            drv_updateScore( );

                            return STATUS_HIT_BRICK;
                        }
                    else
                        min = index;
                else
                    max = index;
            else
                min = index;
        else
            max = index;

        old_index = index;
        index = min + ( max - min ) / 2;
    }

    return STATUS_NO_CONTACT;
}

/* PUBLIC_FUNCTIONS */

/* Initialization */

void demo_generateWall( demo_object_t *wall )
{
    uint16_t x;
    uint16_t y;
    uint8_t cnt;
    uint16_t color;

    cnt = 0;
    color = 0x0000;

    for( y = 0; y < BRICK_MAX_COLUMNS; y++ )
    {
        for( x = 0; x < BRICK_MAX_ROWS; x++ )
        {
            switch( cnt % 3 )
            {
                case 1:
                    color = COLOR_RED;
                    break;
                case 2:
                    color = COLOR_GREEN;
                    break;
                case 0:
                    color = COLOR_BLUE;
            }
            wall[ cnt ].x = x * BRICK_DEF_WIDTH;
            wall[ cnt ].y = y * BRICK_DEF_HEIGHT;
            wall[ cnt++ ].color = color;

            drv_paintObject( wall[ cnt - 1 ], BRICK_DEF_WIDTH, BRICK_DEF_HEIGHT );
        }
    }

    overall_score = 0;
}

void demo_reset( demo_ball_t *ball, demo_object_t *paddle, demo_object_t *wall )
{
    uint16_t cnt = 0;
    char temp_data[ 50 ];

    if( life_cnt == 1 )
    {
        matrixrgb_fillScreen( COLOR_BLACK );

        while( cnt < PANEL_GAME_OVER_SIZE )
        {
            drv_writePixel( game_over[ cnt ], game_over[ cnt + 1 ], COLOR_RED );
            cnt += 2;
        }

        mikrobus_logWrite( " * SCORE: ", _LOG_TEXT );
        IntToStr( overall_score, temp_data );
        Ltrim( temp_data );
        mikrobus_logWrite( temp_data, _LOG_LINE );

        bricks_destroyed = DIF_LEVEL_1;

        Delay_ms( PANEL_PAUSE_LENGTH );
    }
    else if( life_cnt == 0 )
    {
        matrixrgb_fillScreen( COLOR_BLACK );
        Delay_ms( PANEL_PAUSE_LENGTH / 2 );
        demo_generateWall( wall );

        life_cnt = DEF_NUM_LIVES;
    }
    
    life_cnt--;
    
    // Ball
    drv_writePixel( ball->x, ball->y, COLOR_BLACK );

    ball->x = BALL_DEF_X;
    ball->y = BALL_DEF_Y;
    ball->direction = BALL_DEF_DIRECTION;
    drv_writePixel( BALL_DEF_X, BALL_DEF_Y, BALL_DEF_COLOR );

    // Paddle
    paddle->color = COLOR_BLACK;
    drv_paintObject( ( *paddle ), PAD_DEF_WIDTH, PAD_DEF_HEIGHT );

    paddle->x = PAD_DEF_X;
    paddle->y = PAD_DEF_Y;
    paddle->color = PAD_DEF_COLOR;
    drv_paintObject( ( *paddle ), PAD_DEF_WIDTH, PAD_DEF_HEIGHT );
}

/* Application task */

// Move ball
void demo_moveBall( demo_ball_t *ball )
{
    drv_writePixel( ball->x, ball->y, COLOR_BLACK );
    drv_repositionBall( ball );
    drv_writePixel( ball->x, ball->y, BALL_DEF_COLOR );
    
    // Ball speed settings
    switch ( score_multiplier )
    {
        case DIF_MULTIPLIER_1:
            Delay_ms( MOV_SPEED_1 );  
            break;
        case DIF_MULTIPLIER_2:
            Delay_ms( MOV_SPEED_2 );  
            break;
        case DIF_MULTIPLIER_3:
            Delay_ms( MOV_SPEED_3 );  
            break;
        case DIF_MULTIPLIER_4:
            Delay_ms( MOV_SPEED_4 );  
            break;
        default:
            Delay_ms( MOV_SPEED_5 );
    }
}

// Move paddle
void demo_movePaddle( demo_object_t *paddle, uint8_t mode )
{
    uint16_t cmd;

    if( mode == MOV_MODE_SLIDER_2 )
    {
        cmd = ADC1_Read( 3 );
        cmd >>= 8;
        
        if( cmd != old_cmd )
        {
            paddle->color = COLOR_BLACK;
            drv_paintObject( ( *paddle ), PAD_DEF_WIDTH, PAD_DEF_HEIGHT );

            if( ( cmd % 2 ) != 0 )
                paddle->x = ( cmd - 1 ) * 3;
            else
                paddle->x = ( cmd - 2 )  * 3 + 2;

            old_cmd = cmd;

            paddle->color = PAD_DEF_COLOR;
            drv_paintObject( ( *paddle ), PAD_DEF_WIDTH, PAD_DEF_HEIGHT );
        }
    }  
}

// Collision check
uint8_t demo_collisionCheck( demo_ball_t *ball, demo_object_t *paddle, demo_object_t *wall )
{
    uint16_t old_x;
    uint16_t old_y;
    uint8_t status;

    old_x = ball->x;
    old_y = ball->y;

    drv_repositionBall( ball );

    switch( ball->direction )
    {
        case DIR_NORTH:
            if( !( status = drv_edgeCheck( ball ) ) )
                status = drv_brickCheck( ball, wall );
            break;
        case DIR_NORTHEAST:
            if( !( status = drv_edgeCheck( ball ) ) )
                status = drv_brickCheck( ball, wall );
            break;
        case DIR_SOUTHEAST:
            if( !( status = drv_edgeCheck( ball ) ) )
                if( !( status = drv_paddleCheck( ball, paddle ) ) )
                    status = drv_brickCheck( ball, wall );
            break;
        case DIR_SOUTH:
            if( !( status = drv_edgeCheck( ball ) ) )
                if( !( status = drv_paddleCheck( ball, paddle ) ) )
                    status = drv_brickCheck( ball, wall );
            break;
        case DIR_SOUTHWEST:
            if( !( status = drv_edgeCheck( ball ) ) )
                if( !( status = drv_paddleCheck( ball, paddle ) ) )
                    status = drv_brickCheck( ball, wall );
            break;
        case DIR_NORTHWEST:
            if( !( status = drv_edgeCheck( ball ) ) )
                status = drv_brickCheck( ball, wall );
    }

    ball->x = old_x;
    ball->y = old_y;

    if( status == STATUS_HIT_FLOOR )
        demo_reset( ball, paddle, wall );

    return status;
} 